// -*- c++ -*-
#ifndef ARRAY_MAP_H
#define ARRAY_MAP_H

#include <memory>

#include <hugo/map_iterator.h>

///\ingroup graphmaps
///\file
///\brief Graph maps that construates and destruates
///their elements dynamically.

namespace hugo {


  /// \addtogroup graphmaps
  /// @{
	
  /** The ArrayMap template class is graph map structure what
   *  automatically updates the map when a key is added to or erased from
   *  the map. This map factory uses the allocators to implement 
   *  the container functionality.
   *
   *  The template parameter is the MapRegistry that the maps
   *  will belong to and the ValueType.
   */

  template <typename MapRegistry, typename Value> 
  class ArrayMap : public MapRegistry::MapBase {
		
  public:
		
    /// The graph type of the maps. 
    typedef typename MapRegistry::Graph Graph;
    /// The key type of the maps.
    typedef typename MapRegistry::KeyType KeyType;
    /// The iterator to iterate on the keys.
    typedef typename MapRegistry::KeyIt KeyIt;

    /// The MapBase of the Map which imlements the core regisitry function.
    typedef typename MapRegistry::MapBase MapBase;
		
    
  public:

    /// The value type of the map.
    typedef Value ValueType;
    /// The reference type of the map;
    typedef Value& ReferenceType;
    /// The pointer type of the map;
    typedef Value* PointerType;

    /// The const value type of the map.
    typedef const Value ConstValueType;
    /// The const reference type of the map;
    typedef const Value& ConstReferenceType;
    /// The pointer type of the map;
    typedef const Value* ConstPointerType;


    typedef std::allocator<Value> Allocator;

	
    /** Default constructor for the map.
     */
    ArrayMap() : values(0), capacity(0) {}
			
    /** Graph and Registry initialized map constructor.
     */
    ArrayMap(const Graph& g, MapRegistry& r) : MapBase(g, r) {
      allocate_memory();
      for (KeyIt it(*MapBase::getGraph()); it != INVALID; ++it) {
	int id = MapBase::getGraph()->id(it);
	allocator.construct(&(values[id]), Value());
      }								
    }

    /** Constructor to use default value to initialize the map. 
     */
    ArrayMap(const Graph& g, MapRegistry& r, const Value& v) 
      : MapBase(g, r) {
      allocate_memory();
      for (KeyIt it(*MapBase::getGraph()); it != INVALID; ++it) {
	int id = MapBase::getGraph()->id(it);
	allocator.construct(&(values[id]), v);
      }								
    }

    /** Constructor to copy a map of the same map type.
     */
    ArrayMap(const ArrayMap& copy) : MapBase(*copy.graph, *copy.registry) {
      capacity = copy.capacity;
      if (capacity == 0) return;
      values = allocator.allocate(capacity);
      for (KeyIt it(*MapBase::getGraph()); it != INVALID; ++it) {
	int id = MapBase::getGraph()->id(it);
	allocator.construct(&(values[id]), copy.values[id]);
      }
    }

    /** Constructor to copy a map of an other map type.
     */
    template <typename CMap> ArrayMap(const CMap& copy) 
      : MapBase(copy), capacity(0), values(0) {
      if (MapBase::getGraph()) {
	allocate_memory();
	for (KeyIt it(*MapBase::getGraph()); it != INVALID; ++it) {
	  set(it, copy[it]);
	}
      }
    }

    /** Assign operator to copy a map of the same map type.
     */
    ArrayMap& operator=(const ArrayMap& copy) {
      if (&copy == this) return *this;
      if (capacity != 0) {
	MapBase::destroy();
	allocator.deallocate(values, capacity);
      }
      capacity = copy.capacity;
      if (capacity == 0) return *this;
      values = allocator.allocate(capacity);
      for (KeyIt it(*MapBase::getGraph()); it != INVALID; ++it) {
	int id = MapBase::getGraph()->id(it);
	allocator.construct(&(values[id]), copy.values[id]);
      }
      return *this;
    }

    /** Assign operator to copy a map an other map type.
     */
    template <typename CMap> ArrayMap& operator=(const CMap& copy) {
      if (MapBase::getGraph()) {
	MapBase::destroy();
      } 
      MapBase::operator=(copy);
      if (MapBase::getGraph()) {
	allocate_memory();
	for (KeyIt it(*MapBase::getGraph()); it != INVALID; ++it) {
	  set(it, copy[it]);
	}
      }
      return *this;
    }
				
    /** The destructor of the map.
     */
    virtual ~ArrayMap() {
      if (capacity != 0) {
	MapBase::destroy();
	allocator.deallocate(values, capacity);
      }
    }
	
	
    /**
     * The subscript operator. The map can be subscripted by the
     * actual keys of the graph. 
     */
    ReferenceType operator[](const KeyType& key) {
      int id = MapBase::getGraph()->id(key);
      return values[id];
    } 
		
    /**
     * The const subscript operator. The map can be subscripted by the
     * actual keys of the graph. 
     */
    ConstReferenceType operator[](const KeyType& key) const {
      int id = MapBase::getGraph()->id(key);
      return values[id];
    }
	
    /** Setter function of the map. Equivalent with map[key] = val.
     *  This is a compatibility feature with the not dereferable maps.
     */
    void set(const KeyType& key, const ValueType& val) {
      int id = MapBase::getGraph()->id(key);
      values[id] = val;
    }
		
    /** Add a new key to the map. It called by the map registry.
     */
    void add(const KeyType& key) {
      int id = MapBase::getGraph()->id(key);
      if (id >= capacity) {
	int new_capacity = (capacity == 0 ? 1 : capacity);
	while (new_capacity <= id) {
	  new_capacity <<= 1;
	}
	Value* new_values = allocator.allocate(new_capacity);;
	for (KeyIt it(*MapBase::getGraph()); it != INVALID; ++it) {
	  int jd = MapBase::getGraph()->id(it);
	  if (id != jd) {
	    allocator.construct(&(new_values[jd]), values[jd]);
	    allocator.destroy(&(values[jd]));
	  }
	}
	if (capacity != 0) allocator.deallocate(values, capacity);
	values = new_values;
	capacity = new_capacity;
      }
      allocator.construct(&(values[id]), Value());
    }
		
    /** Erase a key from the map. It called by the map registry.
     */
    void erase(const KeyType& key) {
      int id = MapBase::getGraph()->id(key);
      allocator.destroy(&(values[id]));
    }

    /** Clear the data structure.
     */
    void clear() {	
      if (capacity != 0) {
	MapBase::destroy();
	allocator.deallocate(values, capacity);
	capacity = 0;
      }
    }

    /// The stl compatible pair iterator of the map.
    typedef MapIterator<ArrayMap> Iterator;
    /// The stl compatible const pair iterator of the map.
    typedef MapConstIterator<ArrayMap> ConstIterator;

    /** Returns the begin iterator of the map.
     */
    Iterator begin() {
      return Iterator(*this, KeyIt(*MapBase::getGraph()));
    }

    /** Returns the end iterator of the map.
     */
    Iterator end() {
      return Iterator(*this, INVALID);
    }

    /** Returns the begin ConstIterator of the map.
     */
    ConstIterator begin() const {
      return ConstIterator(*this, KeyIt(*MapBase::getGraph()));
    }

    /** Returns the end const_iterator of the map.
     */
    ConstIterator end() const {
      return ConstIterator(*this, INVALID);
    }

  private:
      
    void allocate_memory() {
      int max_id = -1;
      for (KeyIt it(*MapBase::getGraph()); it != INVALID; ++it) {
	int id = MapBase::getGraph()->id(it);
	if (id > max_id) {
	  max_id = id;
	}			
      }
      if (max_id == -1) {
	capacity = 0;
	values = 0;
	return;
      }
      capacity = 1;
      while (capacity <= max_id) {
	capacity <<= 1;
      }
      values = allocator.allocate(capacity);	
    }      

    int capacity;
    Value* values;
    Allocator allocator;
  };		

/// @}

}

#endif
